class KeccakState(object):
    W = 5
    H = 5

    rangeW = range(W)
    rangeH = range(H)

    @staticmethod
    def zero():
        return [[0] * KeccakState.W for x in KeccakState.rangeH]

    @staticmethod
    def format(st):
        rows = []

        def fmt(x):
            return '%016x' % x

        for y in KeccakState.rangeH:
            row = []
            for x in rangeW:
                row.append(fmt(st[x][y]))
            rows.append(' '.join(row))
        return '\n'.join(rows)

    @staticmethod
    def lane2bytes(s, w):
        o = []
        for b in range(0, w, 8):
            o.append((s >> b) & 0xff)
        return o

    @staticmethod
    def bytes2lane(bb):
        r = 0
        for b in reversed(bb):
            r = r << 8 | b
        return r

    def __init__(self, bitrate, b):
        from lunespy.utils.crypto.converters import bits_to_bytes

        self.bitrate = bitrate
        self.b = b

        # only byte-aligned
        assert self.bitrate % 8 == 0
        self.bitrate_bytes = bits_to_bytes(self.bitrate)

        assert self.b % 25 == 0
        self.lanew = self.b // 25

        self.s = KeccakState.zero()

    def __str__(self):
        return KeccakState.format(self.s)

    def absorb(self, bb):
        from lunespy.utils.crypto.converters import bits_to_bytes

        assert len(bb) == self.bitrate_bytes

        bb += [0] * int(bits_to_bytes(self.b - self.bitrate))
        i = 0

        for y in self.rangeH:
            for x in self.rangeW:
                self.s[x][y] ^= KeccakState.bytes2lane(bb[i:i + 8])
                i += 8

    def squeeze(self):
        return self.get_bytes()[:self.bitrate_bytes]

    def get_bytes(self):
        from lunespy.utils.crypto.converters import bits_to_bytes

        out = [0] * int(bits_to_bytes(self.b))
        i = 0
        for y in self.rangeH:
            for x in self.rangeW:
                v = KeccakState.lane2bytes(self.s[x][y], self.lanew)
                out[i:i + 8] = v
                i += 8
        return out

    def set_bytes(self, bb):
        i = 0
        for y in self.rangeH:
            for x in self.rangeW:
                self.s[x][y] = KeccakState.bytes2lane(bb[i:i + 8])
                i += 8


class KeccakSponge(object):
    def __init__(self, bitrate, width, padfn, permfn):
        self.state = KeccakState(bitrate, width)
        self.padfn = padfn
        self.permfn = permfn
        self.buffer = []

    def copy(self):
        from copy import deepcopy

        return deepcopy(self)

    def absorb_block(self, bb):
        self.state.bitrate_bytes = int(self.state.bitrate_bytes)
        assert len(bb) == self.state.bitrate_bytes
        self.state.absorb(bb)
        self.permfn(self.state)

    def absorb(self, s):
        from lunespy.utils.crypto.converters import string_to_list

        self.buffer = string_to_list(s)

        while len(self.buffer) >= self.state.bitrate_bytes:
            self.absorb_block(self.buffer[:self.state.bitrate_bytes])
            self.buffer = self.buffer[self.state.bitrate_bytes:]

    def absorb_final(self):
        padded = self.buffer + self.padfn(len(self.buffer), self.state.bitrate_bytes)
        self.absorb_block(padded)
        self.buffer = []

    def squeeze_once(self):
        rc = self.state.squeeze()
        self.permfn(self.state)
        return rc

    def squeeze(self, l):
        Z = self.squeeze_once()
        while len(Z) < l:
            Z += self.squeeze_once()
        return Z[:int(l)]


class KeccakHash(object):
    def __init__(self):
        from lunespy.utils.crypto.converters import multirate_padding
        from lunespy.utils.crypto.converters import bits_to_bytes
        from lunespy.utils.crypto.converters import keccak_f

        bitrate_bits = 1088
        capacity_bits = 512
        output_bits = 256
        self.sponge = KeccakSponge(
            bitrate_bits,
            bitrate_bits + capacity_bits,
            multirate_padding,
            keccak_f)
        self.digest_size = bits_to_bytes(output_bits)
        self.block_size = bits_to_bytes(bitrate_bits)

    def __repr__(self):
        info = (
            self.sponge.state.bitrate,
            self.sponge.state.b - self.sponge.state.bitrate,
            self.digest_size * 8
        )
        return "<KeccakHash with r=%d, c=%d, image=%d>" % info

    def digest(self, string: str) -> str:
        self.sponge.absorb(string)
        finalised = self.sponge.copy()
        finalised.absorb_final()
        digest = finalised.squeeze(self.digest_size)
        return ''.join(map(chr, digest))





from lunespy.client.wallet import Account


def mount_burn(sender: Account, burn_data: dict) -> dict:
    from lunespy.utils import now
    from lunespy.client.transactions.constants import BurnType
    from lunespy.utils.crypto.converters import sign
    from base58 import b58decode
    import struct

    timestamp: int = burn_data.get('timestamp', now())
    fee: int = burn_data.get('fee', BurnType.fee.value)
    quantity: int = burn_data.get('quantity', 0)
    asset_id: int = burn_data['asset_id']

    bytes_data: bytes = BurnType.to_byte.value + \
        b58decode(sender.public_key) + \
        b58decode(asset_id) + \
        struct.pack(">Q", quantity) + \
        struct.pack(">Q", fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx = {
        "type": BurnType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,

        "assetId": asset_id,
        "quantity": quantity
    }
    return mount_tx


def validate_burn(sender: Account, burn_data: dict) -> bool:
    from lunespy.utils import bcolors

    quantity: int = burn_data.get('quantity', -1)
    asset_id: str = burn_data.get('asset_id', False)

    if not sender.private_key:
        print(bcolors.FAIL + 'Burner `Account` not have a private key' + bcolors.ENDC)
        return False

    if quantity < 0:
        print(bcolors.FAIL + 'Burn `quantity` cannot be less than 0' + bcolors.ENDC)
        return False

    if asset_id == False:
        print(bcolors.FAIL + 'Burn_data `asset_id` cannot exists' + bcolors.ENDC)
        return False

    return True


# todo async
def send_burn(mount_tx: dict, node_url: str) -> dict:
    from requests import post

    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx



from lunespy.client.wallet import Account


def mount_lease(sender: Account, lease_data: dict) -> dict:
    from lunespy.client.transactions.constants import LeaseType
    from lunespy.utils.crypto.converters import sign
    from lunespy.utils import now, lunes_to_unes
    from base58 import b58decode
    import struct

    node_address: str = lease_data['node_address']
    timestamp: int = lease_data.get('timestamp', now())
    amount: int = lunes_to_unes(lease_data['amount'])
    fee: int = lease_data.get('fee', LeaseType.fee.value)

    bytes_data: bytes = LeaseType.to_byte.value + \
        b58decode(sender.public_key) + \
        b58decode(node_address) + \
        struct.pack(">Q", amount) + \
        struct.pack(">Q", fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx: dict = {
        "type": LeaseType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,

        "recipient": node_address,
        "amount": amount
    }
    return mount_tx


def validate_lease(sender: Account, lease_data: dict) -> bool:
    from lunespy.utils import bcolors

    amount: int = lease_data.get('amount', -1)

    if not sender.private_key:
        print(bcolors.FAIL + 'Staker `Account` not have a private key' + bcolors.ENDC)
        return False
    elif amount <= 0:
        print(bcolors.FAIL + 'Leasing `amount` cannot be less than 0' + bcolors.ENDC)
        return False
    return True


# todo async
def send_lease(mount_tx: dict, node_url: str) -> dict:
    from requests import post
    
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx



from requests import get
from enum import Enum


def all_node_conected_in_node_url(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/peers/all'
    else:
        full_url = f'{node_url}/peers/all' # You have pass your node url with https or other contents

    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def node_version(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/utils/lunesnode/version'
    else:
        full_url = f'{node_url}/utils/lunesnode/version' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def version_all_lunes_node_conected(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/peers/connected'
    else:
        full_url = f'{node_url}/peers/connected' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': [
                {
                    'node_url': data['address'][1:],
                    'version': data['applicationVersion']
                }
                for data in response.json()['peers']
            ]
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


class Node(Enum):
    mainnet_url: str = 'https://lunesnode.lunes.io'
    testnet_url: str = 'https://lunesnode-testnet.lunes.io'
    mainnet_blockexplorer: str = 'https://blockexplorer.lunes.io'
    testnet_blockexplorer: str = 'https://blockexplorer-testnet.lunes.io'
    mainnet_total_supply: float = 150_728_537.61498705
    testnet_total_supply: float = 800_100_000.00000000  



from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account
from lunespy.utils import drop_none


class CreateAlias(BaseTransaction):
    def __init__(self, sender: Account, alias: str = None,
                 timestamp: int = None, fee: int = None) -> None:
        self.alias_data: dict = drop_none({
            'timestamp': timestamp,
            'alias': alias,
            'fee': fee
        })
        super().__init__('Alias', self.alias_data)
        self.sender: Account = sender
        self.history: list = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.alias.validators import validate_alias

        return validate_alias(self.sender, self.alias_data)

    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.alias.validators import mount_alias

        return super().transaction(
            mount_tx=mount_alias,
            sender=self.sender,
            alias_data=self.alias_data)

    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.alias.validators import send_alias

        tx = super().send(send_alias, node_url)
        self.history.append(tx)
        return tx


from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account
from lunespy.utils import drop_none


class BurnToken(BaseTransaction):
    def __init__(self, sender: Account, asset_id: str = None, quantity: int = None,
                 timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none
        self.burn_data: dict = drop_none({
            "timestamp": timestamp,
            "asset_id": asset_id,
            "quantity": quantity,
            "fee": fee
        })
        super().__init__('Burn Token', self.burn_data)
        self.sender = sender
        self.history = []

    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.burn.validators import validate_burn

        return validate_burn(self.sender, self.burn_data)

    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.burn.validators import mount_burn

        return super().transaction(
            mount_tx=mount_burn,
            sender=self.sender,
            burn_data=self.burn_data)

    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.burn.validators import send_burn

        tx = super().send(send_burn, node_url)
        self.history.append(tx)
        return tx


class BurnAsset(BurnToken):
    def __init__(self, sender: Account, timestamp: int = None, asset_id: str = None,
                 quantity: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none
        burn_data: dict = drop_none({
            "timestamp": timestamp,
            "asset_id": asset_id,
            "quantity": quantity,
            "fee": fee
        })
        BaseTransaction.__init__(self, tx_type='Burn Asset', tx_data=burn_data)
        BurnToken.__init__(self, sender=sender, **burn_data)


class BurnNFT(BurnAsset):
    def __init__(self, sender: Account, timestamp: int = None, asset_id: str = None,
                 quantity: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none
        burn_data: dict = drop_none({
            "timestamp": timestamp,
            "asset_id": asset_id,
            "quantity": quantity,
            "fee": fee
        })
        BaseTransaction.__init__(self, tx_type='Burn NFT', tx_data=burn_data)
        BurnToken.__init__(self, sender=sender, **burn_data)



from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class CancelLease(BaseTransaction):
    def __init__(self, sender: Account, lease_tx_id: str = None,
                 timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        self.cancel_data = drop_none({
            "timestamp": timestamp,
            "lease_tx_id": lease_tx_id,
            "fee": fee
        })
        super().__init__('Create Lease', self.cancel_data)
        self.sender: Account = sender
        self.history: list = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.cancel_lease.validators import validate_cancel

        return validate_cancel(self.sender, self.cancel_data)


    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.cancel_lease.validators import mount_cancel

        return super().transaction(
            mount_tx=mount_cancel,
            sender=self.sender,
            cancel_data=self.cancel_data)

    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.cancel_lease.validators import send_cancel

        tx = super().send(send_cancel, node_url)
        self.history.append(tx)
        return tx


from lunespy.client.wallet import Account


def mount_mass_transfer(sender: Account, receivers_list: list, mass_transfer_data: dict) -> dict:
    from lunespy.client.transactions.constants import TransferType
    from lunespy.client.transactions.constants import MassType
    from lunespy.utils.crypto.converters import sign
    from lunespy.utils import lunes_to_unes
    from lunespy.utils import now
    from base58 import b58decode
    import struct

    fee: int = TransferType.fee.value + len(receivers_list) * mass_transfer_data.get('fee', MassType.fee.value)
    timestamp: int = mass_transfer_data.get('timestamp', now())
    asset_id: str = mass_transfer_data.get('asset_id', "")
    receivers_list: list = [
        {'receiver': tx['receiver'], 'amount': lunes_to_unes(tx['amount'])}
        for tx in receivers_list
    ]


    hash_transfer: bytes = lambda tx: b58decode(tx['receiver']) + struct.pack(">Q", tx['amount']) 
    receivers_list_data = b''.join(
        map(hash_transfer, receivers_list)
    )

    bytes_data: bytes = MassType.to_byte.value + \
            b'\1' + \
            b58decode(sender.public_key) + \
            (b'\1' + b58decode(asset_id) if asset_id != "" else b'\0') + \
            struct.pack(">H", len(receivers_list)) + \
            receivers_list_data + \
            struct.pack(">Q", timestamp) + \
            struct.pack(">Q", fee)

    signature: bytes = sign(sender.private_key, bytes_data)

    data = {
        "type": MassType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,
        
        "version": 1,
        "assetId": asset_id,
        "transfers": [
            {'recipient': tx['receiver'], 'amount': tx['amount']}
            for tx in receivers_list
        ],
        "proofs": [
            signature.decode()
        ]
    }

    return data


def validate_mass_transfer(sender: Account, receivers_list: list):
    from lunespy.client.wallet.validators import validate_address
    from lunespy.utils import bcolors

    amounts_list: list = [tx['amount'] for tx in receivers_list]
    
    if not sender.private_key:
        print(bcolors.FAIL + 'Sender `Account` not have a private key' + bcolors.ENDC)
        return False
    elif not all(filter(lambda address: validate_address(address['receiver'], sender.chain_id), receivers_list)):
        return False
    elif len(receivers_list) > 100:
        print(bcolors.FAIL + '`receivers_list` cannot contain more than 100 addresses' + bcolors.ENDC)
        return False
    elif not all(filter(lambda amount: amount <= 0, amounts_list)):
        print(bcolors.FAIL + 'Some address contains 0 as `amount`' + bcolors.ENDC)
        return False
    return True


def send_mass_transfer(mount_tx: dict, node_url: str) -> dict:
    from requests import post

    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx



from lunespy.client.wallet import Account


def validate_alias(sender: Account, alias_data: dict) -> bool:
    from lunespy.utils import bcolors

    alias: int = alias_data.get('alias', False)
    valid_alias_characters =  "-.0123456789@_abcdefghijklmnopqrstuvwxyz"

    if not sender.private_key:
        print(bcolors.FAIL + 'Sender `Account` not have a private key' + bcolors.ENDC)
        return False

    if alias == False:
        print(bcolors.FAIL + 'Alias_data `alias` dont exists' + bcolors.ENDC)
        return False
    
    if not all(each_char in valid_alias_characters for each_char in alias):
        print(
            bcolors.FAIL + \
            "`Alias` should contain only following characters: -.0123456789@_abcdefghijklmnopqrstuvwxyz" + \
            bcolors.ENDC)
        return False
    
    
    return True


def mount_alias(sender: Account, alias_data: dict) -> dict:
    from lunespy.utils import now
    from lunespy.utils.crypto.converters import sign
    from lunespy.client.transactions.constants import AliasType
    from lunespy.utils.crypto.converters import string_to_bytes
    from base58 import b58decode
    import struct

    timestamp: int = alias_data.get('timestamp', now())
    fee: int = alias_data.get('fee', AliasType.fee.value)
    alias: str = alias_data['alias']
    chain_id: str = sender.chain_id

    aliasWithNetwork = AliasType.mount.value +\
        string_to_bytes(str(chain_id)) + \
        struct.pack(">H", len(alias)) + \
        string_to_bytes(alias)

    bytes_data = AliasType.to_byte.value + \
        b58decode(sender.public_key) + \
        struct.pack(">H", len(aliasWithNetwork)) + \
        aliasWithNetwork + \
        struct.pack(">Q", fee) + \
        struct.pack(">Q", timestamp)

    signature: bytes = sign(sender.private_key, bytes_data)

    mount_tx = {
        "type": AliasType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,

        "alias": alias
    }
    return mount_tx


def send_alias(mount_tx: dict, node_url: str) -> dict:
    from requests import post
    
    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx



class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def flat_map(listOfLists):
    from itertools import chain
    return list(chain.from_iterable(listOfLists))


def log_data(data: dict) -> None:
    list(map(
        lambda tuple: print(
            f"{tuple[0]}{bcolors.OKGREEN} ─> {str(tuple[1])}{bcolors.ENDC}"
        ),
        data.items()
        )
    )


def export_json(data: dict, name: str, path: str) -> bool:
    import json
    path = path.replace("//", "/")
    full_path = f"{path}{name}.json"
    try:
        with open(full_path, 'w') as file:
            file.write( json.dumps(data) )
    except Exception as msg:
        raise Exception(
            bcolors.FAIL + f"[Error] File Don't Saved Because:\n└──{msg}" + bcolors.ENDC
        )

    return f"file save in {full_path}"


def semantic_version() -> str:
    from subprocess import check_output

    def get_logs() -> list[str]:
        return check_output(
            'git log --pretty="%s"',
            shell=True
        ).decode().split('\n')

    major, minor, patch = 0,0,0

    for commit in get_logs():
        if commit.startswith("deprecated"):
            patch = 0
            minor = 0
            major += 1
        elif  commit.startswith("Merge" or "issued" or "merged"):
            patch = 0
            minor += 1
        elif  commit.startswith("fixed" or "fix" or "Update"):
            patch += 1
    print(
        bcolors.OKGREEN + f"v{major}.{minor}.{patch}" + bcolors.ENDC
    )
    return f"v{major}.{minor}.{patch}"


def changelog():
    from subprocess import check_output
    from os import system

    deprecated = ["## Deprecated"]
    merged_issued = ["## Issued"]
    fixed = ["## Fixed"]
    refactored = ["## Refactored"]
    removed = ["## Removed"]
    other = ["## Others"]
    changelog = [
        [f"# Changelog {semantic_version()}"],
        deprecated, merged_issued, fixed, refactored, removed, other
    ]

    def get_logs() -> list[str]:
        return check_output(
            'git log --pretty="- [%h](%H) %s"',
            shell=True
        ).decode().split('\n')

    for commit in get_logs():
        if len(commit.split(" ")) > 3:
            test = commit.split(" ")[2]

            if test.startswith("deprecated"):
                deprecated.append(commit)
            elif  test.startswith("Merge" or "issued" or "merged" or "merg"):
                merged_issued.append(commit)
            elif  test.startswith("fixed" or "fix" or "Update"):
                fixed.append(commit)
            elif  test.startswith("refact" or "chang"):
                refactored.append(commit)
            elif  test.startswith("remov"):
                removed.append(commit)
        else:
            other.append(commit)

    with open('./CHANGELOG.md', 'w') as file:
        for line in flat_map(changelog):
            file.write(line + "\n")


def now() -> int:
    from time import time

    return int(
        time() * 1000
    )


def lunes_to_unes(lunes: float or int) -> int:
    return int(lunes * 10e7)


def unes_to_lunes(unes: int) -> float:
    return float(unes / 10e7)


def sha256(object: object) -> str:
    from hashlib import sha256

    return sha256(
        str(object).encode()
    ).hexdigest()


def drop_none(data: dict) -> dict:
    validate_keys = list(filter(
        lambda key: data[key] != None,
        data.keys()
    ))

    return {
        key: data[key]
        for key in data.keys()
        if key in validate_keys
    }



from requests import get
from lunespy.server.nodes import Node

def transaction_from_id(id: str) -> dict:
    full_url = f'http://lunesnode-testnet.lunes.io/transactions/info/{id}'
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else: 
        return {
            'status': 'error',
            'response': response.text
        }


def unconfirmed_transaction(node_url: str) -> dict:
    full_url = f'{node_url}/transactions/unconfirmed' # You have pass your node url with https or other contents
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'stauts': 'error',
            'response': response.text
        }


def transactions_from_address(address: str, limit_transactions: int, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/transactions/address/{address}/limit/{limit_transactions}'
    else:
        full_url = f'{node_url}/transactions/address/{address}/limit/{limit_transactions}' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else: 
        return {
            'status': 'error',
            'response': response.text
        }



from lunespy.server.nodes import Node
from lunespy.utils import unes_to_lunes
from lunespy.utils import export_json
from requests import get


def address_associated_with_an_alias(alias: str, node_url: str = None) -> dict:
    if node_url == None:
       full_url = f'{Node.mainnet_url.value}/addresses/alias/by-alias/{alias}'

    else:
        full_url = f'{node_url}/addresses/alias/by-alias/{alias}' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def asset_distribution(asset_id: str, node_url: str == None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/assets/{asset_id}/distribution'
    else: 
        full_url = f'{node_url}/assets/{asset_id}/distribution' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def balance_all_assets_of_address(address: str, node_url: str == None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/assets/balance/{address}'
    else:
        full_url = f'{node_url}/assets/balance/{address}' # You have pass your node url with https or other contents
    
    response = get(full_url)
    
    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def balance_for_especify_asset_of_address(address: str, asset_id: str, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/assets/balance/{address}/{asset_id}'
    else:
        full_url = f'{node_url}/assets/balance/{address}/{asset_id}' # You have pass your node url with https or other contents 
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def balance_of_all_address(node_ip: str, node_api_key: str) -> dict:    
    full_url = f"{node_ip}/debug/state" # You have pass your node url with https or other contents
    header = {"X-API-key": node_api_key}
    response = get(full_url, headers=header)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def balance_of_address(address: str, node_url: str = None) -> int:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/addresses/balance/{address}'
    else:
        full_url = f'{node_url}/addresses/balance/{address}' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def list_of_rich(**kargs: dict) -> dict:
    """
    Example:
        quantity=30,
        node_ip_port="127.0.0.1:555",
        net="mainnet" or "testnet"
        node_api_key="",
        export=True or False,
        path='./data/'
    """
    def percent(amount: float) -> float:
        return round((amount / supply) * 100, 5)

    def percent_total(wallets: list) -> float:
        total = sum([
            i['amount']
            for i in wallets
        ])
        return percent(total)
    
    response = balance_of_all_address(kargs['node_ip'], kargs['node_api_key'])
    supply = Node.mainnet_total_supply.value if kargs['net'] == 'mainnet' else Node.testnet_total_supply.value
    link = Node.mainnet_blockexplorer.value if kargs['net'] == 'mainnet' else Node.testnet_blockexplorer.value

    if response['status'] != 'ok':
        return  {
            'status': 'error',
            'response': response['response']
        }
    else:
        wallets_lunes = dict(zip(
            response['response'],
            list(map(
                lambda item: unes_to_lunes(item),
                response['response'].values()
            ))
        ))

        wallets = sorted(
            [
                {
                    'address': address,
                    'amount': amount,
                    'percent': percent(amount),
                    'link': f'{link}/address/{address}'
                }
                for address, amount in list(
                    wallets_lunes.items()
                )
            ],
            key = lambda item: item.get('amount'),
            reverse = True
        )[:kargs.get('quantity')]

        report = {
            "total_supply": supply,
            "chain": kargs['net'],
            "total_percent": percent_total(wallets),
            "wallet_list": wallets
        }

        if kargs.get('export', False):
            export_json(
                report,
                "rich_list",
                kargs.get('path', './data/')
            )

        return  {
            'status': 'ok',
            'response': report
        }


def leasing_active_by_address(address: str, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/leasing/active/{address}'
    else:
        full_url = f'{node_url}/leasing/active/{address}' # You have pass your node url with https or other contents

    response = get(full_url)
    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': 
                {
                    full_url.replace('/addresses',''): response.json()
                }
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def address_of_node_from_url(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/addresses'

    else:
        full_url = f'{node_url}/addresses' # You have pass your node url with https or other contents
            
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def aliases_associated_with_an_address(address: str, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}addresses/alias/by-address/{address}'
    else:
        full_url = f'{node_url}/addresses/alias/by-address/{address}' # You have pass your node url with https or other contents

    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else: 
        return {
            'status': 'error',
            'response': response.text
        }



def to_human_b58(data: bytes) -> str:
    from base58 import b58encode

    return b58encode(data).decode()


def to_machine_b58(data: str) -> bytes:
    from base58 import b58decode

    return b58decode(data)


def validate_sign(public_key, message, signature) -> bool:
    from axolotl_curve25519 import verifySignature

    verified = verifySignature(
        to_machine_b58(public_key),
        to_machine_b58(message),
        to_machine_b58(signature)
    )

    return True if verified == 0 else False


def multirate_padding(used_bytes, align_bytes):
    padlen = align_bytes - used_bytes
    if padlen == 0:
        padlen = align_bytes
    # note: padding done in 'internal bit ordering', wherein LSB is leftmost
    if padlen == 1:
        return [0x81]
    else:
        return [0x01] + ([0x00] * (int(padlen) - 2)) + [0x80]


def bits_to_bytes(x):
    return (int(x) + 7) / 8


def string_to_bytes(string: str) -> bytes:
    return string.encode('latin-1')


def string_to_list(string: str) -> list:
    return [char for char in string]


def bytes_to_string(bytes: bytes, decode: bool=False) -> str:
    if not decode:
        return ''.join(
            map(chr, bytes)
        )
    else:
        from base58 import b58decode

        return ''.join(
            map(chr, b58decode(bytes))
        )


def hash_data(data: str) -> str:
    from lunespy.utils.crypto.algorithms import KeccakHash
    from hashlib import blake2b
    keccak256 = KeccakHash()

    return keccak256.digest( 
        blake2b(data, digest_size=32).digest()
    )


def sign(private_key: str, message: bytes) -> bytes:
    from axolotl_curve25519 import calculateSignature as curve
    from os import urandom

    return curve(
            urandom(64),
            to_machine_b58(private_key),
            message
        )


def sha256(string: str) -> str:
    from hashlib import sha256

    return sha256(
        string_to_bytes(string)
    ).digest()


def ror(value, right, bits):
    from lunespy.utils.crypto import Masks

    top = value >> right
    bot = (value & Masks[right]) << (bits - right)
    return bot | top


def rol(value, left, bits):
    from lunespy.utils.crypto import Masks

    top = value >> (bits - left)
    bot = (value & Masks[bits - left]) << left
    return bot | top


def keccak_f(state):
    from lunespy.utils.crypto import RoundConstants
    from math import log

    def round(A, RC):
        from lunespy.utils.crypto import RotationConstants
        from functools import reduce
        from operator import xor

        W, H = state.W, state.H
        rangeW, rangeH = state.rangeW, state.rangeH
        lanew = state.lanew
        zero = state.zero

        # theta
        C = [reduce(xor, A[x]) for x in rangeW]
        D = [0] * W
        for x in rangeW:
            D[x] = C[(x - 1) % W] ^ rol(C[(x + 1) % W], 1, lanew)
            for y in rangeH:
                A[x][y] ^= D[x]

        # rho and pi
        B = zero()
        for x in rangeW:
            for y in rangeH:
                B[y % W][(2 * x + 3 * y) % H] = rol(A[x][y], RotationConstants[y][x], lanew)

        # chi
        for x in rangeW:
            for y in rangeH:
                A[x][y] = B[x][y] ^ ((~ B[(x + 1) % W][y]) & B[(x + 2) % W][y])

        # iota
        A[0][0] ^= RC

    l = int(log(state.lanew, 2))
    nr = 12 + 2 * l

    for ir in range(nr):
        round(state.s, RoundConstants[ir])



from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class ReissueToken(BaseTransaction):
    def __init__(self, sender: Account, asset_id: str = None, reissuable: bool = None,
                 quantity: int = None, timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        self.reissue_data = drop_none({
            "asset_id": asset_id,
            "reissuable": reissuable,
            "quantity": quantity,
            "timestamp": timestamp,
            "fee": fee
        })
        super().__init__('Reissue Token', self.reissue_data)
        self.sender = sender
        self.history = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.reissue.validators import validate_reissue

        return validate_reissue(self.sender, self.reissue_data)


    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.reissue.validators import mount_reissue

        return super().transaction(
            mount_tx=mount_reissue,
            sender=self.sender,
            reissue_data=self.reissue_data)



    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.reissue.validators import send_reissue

        tx = super().send(send_reissue, node_url)
        self.history.append(tx)
        return tx


class ReissueAsset(ReissueToken):
    def __init__(self, sender: Account, asset_id: str = None, reissuable: bool = None,
                 quantity: int = None, timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        reissue_data = drop_none({
            "asset_id": asset_id,
            "reissuable": reissuable,
            "quantity": quantity,
            "timestamp": timestamp,
            "fee": fee
        })
        BaseTransaction.__init__(self, tx_type='Reissue Asset', tx_data=reissue_data)
        ReissueToken.__init__(self, sender=sender, **reissue_data)


class ReissueNFT(ReissueToken):
    def __init__(self, sender: Account, asset_id: str = None, reissuable: bool = None,
                 quantity: int = None, timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        reissue_data = drop_none({
            "asset_id": asset_id,
            "reissuable": reissuable,
            "quantity": quantity,
            "timestamp": timestamp,
            "fee": fee
        })
        BaseTransaction.__init__(self, tx_type='Reissue NFT', tx_data=reissue_data)
        ReissueToken.__init__(self, sender=sender, **reissue_data)



def validate_wallet(wallet: dict) -> dict:
    from lunespy.client.wallet.errors import InvalidNonce
    from lunespy.client.wallet.generators import wallet_generator

    wallet['nonce'] = wallet.get('nonce', 0)

    if wallet.get('chain', 'mainnet') == 'mainnet':
        wallet['chain_id'] = '1'
        wallet['chain'] = 'mainnet'
    else:
        wallet['chain_id'] = '0'
        wallet['chain'] = 'testnet'

    if wallet['nonce'] not in range(0, 4_294_967_295 + 1):
        raise InvalidNonce

    elif wallet.get('seed', False):
        return wallet_generator(**wallet)

    elif wallet.get('private_key', False):
        return wallet_generator(**wallet)

    elif wallet.get('public_key', False):
        return wallet_generator(**wallet)

    elif wallet.get('address', False):
        if validate_address(wallet['address'], chain_id=wallet['chain_id']):
            return {
                'private_key': '',
                'public_key': '',
                'address': wallet['address'],
                'nonce': 0,
                'chain': wallet['chain'],
                'chain_id': wallet['chain_id'],
                'seed': '',
                'hash_seed': '',
                'byte_private_key': b'',
                'byte_public_key': b'',
                'byte_address': wallet['address'].encode()
            }

    elif wallet.get('alias'):
        from lunespy.server.address import aliases

        return {
            'private_key': '',
            'public_key': '',
            'address': aliases(wallet['alias']),
            'nonce': 0,
            'seed': '',
            'byte_private_key': b'',
            'byte_public_key': b'',
            'byte_address': aliases(wallet['alias']).encode(),
        }

    else:
        return wallet_generator(**wallet)


def validate_address(address: str, chain_id: str) -> bool:
    from lunespy.client.wallet.errors import InvalidChecksumAddress
    from lunespy.client.wallet.errors import InvalidVersionAddress
    from lunespy.client.wallet.errors import InvalidLengthAddress
    from lunespy.client.wallet.errors import InvalidChainAddress
    from lunespy.client.wallet.constants import ADDRESS_CHECKSUM_LENGTH
    from lunespy.client.wallet.constants import ADDRESS_VERSION
    from lunespy.client.wallet.constants import ADDRESS_LENGTH
    from lunespy.utils.crypto.converters import bytes_to_string
    from lunespy.utils.crypto.converters import string_to_bytes
    from lunespy.utils.crypto.converters import hash_data

    bytes_address = bytes_to_string(address, decode=True)
    checksum = bytes_address[-ADDRESS_CHECKSUM_LENGTH:]
    length_checksum = bytes_address[:-ADDRESS_CHECKSUM_LENGTH]
    length_checksum_bytes = string_to_bytes(length_checksum)
    hash_checksum = hash_data(length_checksum_bytes)
    chain = hash_checksum[:ADDRESS_CHECKSUM_LENGTH]

    if checksum != chain:
        raise InvalidChecksumAddress
    elif bytes_address[0] != chr(ADDRESS_VERSION):
        raise InvalidVersionAddress
    elif bytes_address[1] != chain_id:
        raise InvalidChainAddress
    elif len(bytes_address) != ADDRESS_LENGTH:
        raise InvalidLengthAddress
    return True



from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class MassTransferToken(BaseTransaction):
    """
    receivers_list: list
        tx: dict
            receiver: str
            amount: float

    transfer_data: dict
        timestamp: int
        asset_id: str
    """
    def __init__(self, sender: Account, receivers_list: list, **mass_transfer_data: dict) -> None:
        super().__init__("Mass", {'receivers_list': receivers_list, 'mass_transfer_data': mass_transfer_data})
        self.sender: Account = sender
        self.receivers_list: list[dict] = receivers_list
        self.mass_transfer_data: dict = mass_transfer_data
        self.history: list = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.mass.validators import validate_mass_transfer

        return validate_mass_transfer(self.sender, self.receivers_list)


    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.mass.validators import mount_mass_transfer

        return super().transaction(
            mount_tx=mount_mass_transfer,
            sender=self.sender,
            receivers_list=self.receivers_list,
            mass_transfer_data=self.mass_transfer_data)


    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.mass.validators import send_mass_transfer

        tx = super().send(send_mass_transfer, node_url)
        self.history.append(tx)
        return tx



from lunespy.client.transactions import BaseTransaction

class TransferToken(BaseTransaction):
    def __init__(self, sender: str, receiver: str, chain: str, amount: float, asset_fee: int=None, asset_id: str=None, timestamp: int=None, fee: int=None) -> None:

        from lunespy.client.transactions.constants import TransferType
        from lunespy.utils import now, lunes_to_unes

        self.timestamp: int = timestamp if timestamp != None else now()
        self.fee: int = fee if fee != None else TransferType.fee.value
        self.asset_id: str = asset_id if asset_id != None else ""
        self.asset_fee: str = asset_fee if asset_fee != None else ""
        if self.asset_id == "":
            self.amount: int = lunes_to_unes(amount)
        else:
            self.amount: int = amount
        self.receiver: str = receiver
        self.sender: str = sender
        self.chain: str = chain
        self.chain_id: str = "1" if chain == "mainnet" else "0"
        self.history: list = []
        self._tx = None

        super().__init__('Transfer', self.__dict__)


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.transfer.validators import validate_transfer

        return validate_transfer(
            sender=self.sender,
            receiver=self.receiver,
            amount=self.amount,
            chain=self.chain
        )


    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.transfer.validators import mount_transfer

        return super().transaction(
            mount_tx=mount_transfer,
            timestamp=self.timestamp,
            asset_fee=self.asset_fee,
            receiver=self.receiver,
            asset_id=self.asset_id,
            amount=self.amount,
            sender=self.sender,
            chain_id=self.chain_id,
            fee=self.fee
        ) if not self._tx else self._tx


    def sign(self, private_key: str) -> dict:
        from lunespy.client.transactions.transfer.validators import sign_transaction

        self._tx = super().sign(
            sign_tx=sign_transaction,
            private_key=private_key,
            **self.transaction
        )
        return self._tx


    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.transfer.validators import send_transfer

        tx = super().send(send_transfer, node_url)
        self.history.append(tx)
        return tx



from lunespy.utils import export_json, log_data, sha256
from lunespy.utils import bcolors
from abc import abstractmethod
from abc import ABCMeta


class BaseTransaction(metaclass=ABCMeta):
    def __init__(self, tx_type: str, tx_data: dict):
        self.tx_type: str = tx_type
        self.tx_data: str = tx_data


    @abstractmethod
    def ready(self) -> bool:
        raise NotImplementedError


    def transaction(self, mount_tx, **extra) -> dict:
        if self.ready:
            tx = {'ready': True}
            tx.update(mount_tx( **extra ))
            return tx
        else:
            print(bcolors.FAIL + f'{self.tx_type} Transactions bad formed', bcolors.ENDC)
            return {'ready': False}


    def sign(self, sign_tx, private_key, **tx) -> dict:
        return sign_tx(private_key, **tx)


    def send(self, send_tx, node_url: str, chain: str) -> dict:
        from lunespy.server.nodes import Node

        if node_url == None:
            if self.sender.chain == chain:
                node_url = Node.mainnet_url.value
            else:
                node_url = Node.testnet_url.value


        if self._tx['ready']:
            tx_response = send_tx(self._tx, node_url=node_url)

            if tx_response['send']:
                id = tx_response['response'].get('id', sha256(tx_response))
                msg = f'tx-{self.tx_type.replace(" ", "-")}-{id}'
                self.show(
                    name=msg,
                    data=tx_response
                )
                return tx_response
            else:
                print(bcolors.FAIL + f"Your {self.tx_type} dont sended because:\n└──{tx_response['response']}" + bcolors.ENDC)
                return tx_response

        else:
            print(bcolors.FAIL + f'{self.tx_type} Transaction dont send', bcolors.ENDC)
            return self._tx


    def show(self, name: str, data: dict, path: str = './') -> None:
        log_data(data)
        export_json(data, name, path)



from lunespy.utils.crypto.converters import bytes_to_string


def validate_transfer(sender: str, receiver: str, amount: int, chain: str) -> bool:
    from lunespy.client.wallet.validators import validate_address
    from lunespy.utils import bcolors
    from base58 import alphabet


    if amount <= 0:
        print(bcolors.FAIL + 'Amount dont should be more than 0' + bcolors.ENDC)
        return False
    elif not all([i in alphabet.decode() for i in sender]):
        print(bcolors.FAIL + 'Sender invalid `public key`' + bcolors.ENDC)
        return False
    elif not validate_address(receiver, "1" if chain == "mainnet" else "0"):
        return False
    else:
        return True


def mount_transfer(sender: str, timestamp: str, receiver: str, asset_fee: str, asset_id: str, amount: int, chain_id: str, fee: int) -> dict:
    from lunespy.client.transactions.constants import TransferType
    from lunespy.utils.crypto.converters import to_machine_b58
    from lunespy.client.wallet.generators import address_generator

    address = address_generator(to_machine_b58(sender), chain_id)['address']
    return {
        "type": TransferType.to_int.value,
        "senderPublicKey": sender,
        "timestamp": timestamp,
        "recipient": receiver,
        "feeAsset": asset_fee,
        "assetId": asset_id,
        "amount": amount,
        "sender": address,
        "fee": fee
    }


def sign_transaction(private_key: str, **tx: dict) -> dict:
    from lunespy.utils.crypto.converters import to_machine_b58, to_human_b58
    from lunespy.client.transactions.constants import TransferType
    from lunespy.utils.crypto.converters import sign
    from struct import pack


    bytes_data: bytes = TransferType.to_byte.value + \
        to_machine_b58(tx["senderPublicKey"]) + \
        (b'\1' + to_machine_b58(tx["assetId"]) if tx["assetId"] != "" else b'\0') + \
        (b'\1' + to_machine_b58(tx["feeAsset"]) if tx["feeAsset"] != "" else b'\0') + \
        pack(">Q", tx["timestamp"]) + \
        pack(">Q", tx["amount"]) + \
        pack(">Q", tx["fee"]) + \
        to_machine_b58(tx["recipient"])

    tx["signature"] = to_human_b58(sign(private_key, bytes_data))
    tx["rawData"] = to_human_b58(bytes_data)
    return tx


# todo async
def send_transfer(mount_tx: dict, node_url: str) -> dict:
    from requests import post

    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx



from lunespy.utils.crypto.converters import to_human_b58, to_machine_b58
from lunespy.utils.crypto.converters import string_to_bytes
from lunespy.utils.crypto.converters import hash_data
from lunespy.client.wallet.constants import word_list
from os import urandom


def address_generator(public_key: bytes, chain_id: str) -> dict:
    un_hashed_address = chr(1) + str(chain_id) + hash_data(public_key)[0:20]
    address_hash = hash_data(string_to_bytes(un_hashed_address))[0:4]
    address = to_human_b58(string_to_bytes(un_hashed_address + address_hash))
    public_key_b58 = to_human_b58(public_key)
    return {
        'seed': "",
        'hash_seed': "",
        'nonce': 0,
        'chain': 'mainnet' if chain_id == '1' else 'testnet',
        'chain_id': chain_id,
        'private_key': "",
        'public_key': public_key_b58,
        'address': address
    }


def new_seed_generator(n_words: int) -> str:
    from lunespy.utils.crypto.converters import bytes_to_string

    def f():
        wordCount = 2048
        r = bytes_to_string(urandom(4))
        x = (ord(r[3])) + (ord(r[2]) << 8) + (ord(r[1]) << 16) + (ord(r[0]) << 24)
        w1 = x % wordCount
        w2 = ((int(x / wordCount) >> 0) + w1) % wordCount
        w3 = ((int((int(x / wordCount) >> 0) / wordCount) >> 0) + w2) % wordCount
        return w1, w2, w3

    n_words_multiple_of_3: int = n_words // 3

    return " ".join([
        word_list[n]
        for _ in range(n_words_multiple_of_3)
        for n in f()
    ])


def seed_generator(seed: str, nonce: int, chain_id: str) -> dict:
    from lunespy.utils.crypto.converters import sha256
    from axolotl_curve25519 import generatePrivateKey, generatePublicKey
    from struct import pack

    hash_seed = hash_data(
        pack(">L", nonce) + string_to_bytes(seed)
    )
    account_hash_seed = sha256(hash_seed)
    private_key = generatePrivateKey(account_hash_seed)
    public_key = generatePublicKey(private_key)
    address = address_generator(public_key, chain_id)
    return {
        'seed': seed,
        'hash_seed': to_human_b58(seed),
        'nonce': nonce,
        'chain': 'mainnet' if chain_id == '1' else 'testnet',
        'chain_id': chain_id,
        'private_key': to_human_b58(private_key),
        'public_key': to_human_b58(public_key),
        'address': address['address']
    }


def private_key_generator(private_key: str, chain_id: str) -> dict:
    from axolotl_curve25519 import generatePublicKey

    private_key_b58 = to_machine_b58(private_key)
    public_key = generatePublicKey(private_key_b58)
    address = address_generator(public_key, chain_id)
    return {
        'seed': "",
        'hash_seed': "",
        'nonce': 0,
        'chain': 'mainnet' if chain_id == '1' else 'testnet',
        'chain_id': chain_id,
        'private_key': to_human_b58(private_key_b58),
        'public_key': to_human_b58(public_key),
        'address': address['address']
    }


def public_key_generator(public_key: str, chain_id: str) -> dict:
    public_key_b58 = to_machine_b58(public_key)
    address = address_generator(public_key_b58, chain_id)
    return {
        'seed': "",
        'hash_seed': "",
        'nonce': 0,
        'chain': 'mainnet' if chain_id == '1' else 'testnet',
        'chain_id': chain_id,
        'private_key': "",
        'public_key': to_human_b58(public_key_b58),
        'address': address['address']
    }


def wallet_generator(**data: dict) -> dict:
    if data.get('seed', False):
        return seed_generator(seed=data['seed'], nonce=data['nonce'], chain_id=data['chain_id'])

    elif data.get('private_key', False):
        return private_key_generator(private_key=data['private_key'], chain_id=data['chain_id'])

    elif data.get('public_key', False):
        return public_key_generator(public_key=data['public_key'], chain_id=data['chain_id'])

    elif data.get('address', False):
        return address_generator(public_key="", chain_id=data['chain_id'])

    elif data.get('n_words', False):
        seed = new_seed_generator(n_words=data['n_words'])
        return seed_generator(seed=seed, nonce=data['nonce'], chain_id=data['chain_id'])

    else:
        n_words: int = 12
        seed: str = new_seed_generator(n_words)
        return seed_generator(seed=seed, nonce=data['nonce'], chain_id=data['chain_id'])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lunespy',
 'lunespy.client',
 'lunespy.client.transactions',
 'lunespy.client.transactions.alias',
 'lunespy.client.transactions.burn',
 'lunespy.client.transactions.cancel_lease',
 'lunespy.client.transactions.issue',
 'lunespy.client.transactions.lease',
 'lunespy.client.transactions.mass',
 'lunespy.client.transactions.reissue',
 'lunespy.client.transactions.transfer',
 'lunespy.client.wallet',
 'lunespy.server.address',
 'lunespy.server.blocks',
 'lunespy.server.nodes',
 'lunespy.server.transactions',
 'lunespy.utils',
 'lunespy.utils.crypto']

package_data = \
{'': ['*'],
 'lunespy': ['.git/*',
             '.git/hooks/*',
             '.git/info/*',
             '.git/logs/*',
             '.git/logs/refs/heads/*',
             '.git/objects/pack/*',
             '.git/refs/heads/*']}

install_requires = \
['base58>=2.1.0,<3.0.0',
 'pytest-parallel>=0.1.1,<0.2.0',
 'python-axolotl-curve25519>=0.4.1.post2,<0.5.0',
 'requests>=2.26.0,<3.0.0']

entry_points = \
{'console_scripts': ['log = lunespy.utils:changelog',
                     'semver = lunespy.utils:semantic_version']}

setup_kwargs = {
    'name': 'lunespy',
    'version': '2.3.0a0',
    'description': 'Library for communication with nodes in mainnet or testnet of the lunes-blockchain network Allows the automation of sending assets, issue end reissue tokens, leasing, registry, and create new wallet.',
    'long_description': "# LunesPy\n\n**The [old version](https://github.com/lunes-platform/lunespy/blob/main/deprecated) is being discontinued, but it can still be used at your own and risk.**\n\nLibrary for communication with nodes in mainnet or testnet of the lunes-blockchain network\nAllows the automation of **sending assets**, **issue end reissue tokens**, **lease** and **create new wallet**.\n\n\n## [What's new?](https://github.com/lunes-platform/lunespy/blob/main/CHANGELOG.md)\n\n\n## [How to use LunesPy?](https://blockchain.lunes.io/telescope/docs/sdk/getting_started)\n\n\n## [Want to contribute to LunesPy?](https://github.com/lunes-platform/lunespy/blob/main/CONTRIBUTING.md)\n\n\n## Contributors\n\nThanks to the following people who have contributed to this project:\n\n* [olivmath](https://github.com/olivmath)\n* [marcoslkz](https://github.com/marcoslkz)\n* [VanJustin](https://github.com/VanJustin)\n* [xonfps](https://github.com/xonfps)\n\n## Contact\n\nIf you want to contact me you can reach me at <development@lunes.io>.\n\n## License\n\n[Apache License Version 2.0](https://github.com/lunes-platform/lunespy/blob/main/LICENSE).\n",
    'author': 'Lunes Platform',
    'author_email': 'development@lunes.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from lunespy.client.wallet import Account


def mount_cancel(sender: Account, cancel_data: dict) -> dict:
    from lunespy.client.transactions.constants import CancelLeaseType
    from lunespy.utils.crypto.converters import sign
    from lunespy.utils import now
    from base58 import b58decode
    import struct

    lease_tx_id: str = cancel_data['lease_tx_id']
    timestamp: int = cancel_data.get('timestamp', now())
    fee: int = cancel_data.get('fee', CancelLeaseType.fee.value)

    bytes_data: bytes = CancelLeaseType.to_byte.value + \
        b58decode(sender.public_key) + \
        struct.pack(">Q", fee) + \
        struct.pack(">Q", timestamp) + \
        b58decode(lease_tx_id)

    signature: bytes = sign(sender.private_key, bytes_data)
    mount_tx: dict = {
        "type":CancelLeaseType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,

        "leaseId": lease_tx_id
    }
    return mount_tx


def validate_cancel(sender: Account, cancel_data: dict) -> bool:
    from lunespy.utils import bcolors

    amount: int = cancel_data.get('amount', -1)
    lease_tx_id: str = cancel_data.get('lease_tx_id', '')

    if not sender.private_key:
        print(bcolors.FAIL + 'Staker `Account` not have a private key' + bcolors.ENDC)
        return False
    elif lease_tx_id == '':
        print(bcolors.FAIL + 'Leasing must be `lease_tx_id`' + bcolors.ENDC)
        return False
    return True


# todo async
def send_cancel(mount_tx: dict, node_url: str) -> dict:
    from requests import post

    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })

    if response.status_code in range(200, 300):
        mount_tx['send'] = True
        mount_tx['response'] = response.json()
        return mount_tx
    else:
        mount_tx['send'] = False
        mount_tx['response'] = response.text
        return mount_tx


from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account


class IssueToken(BaseTransaction):
    """
    issue_data: dict
        description: str
        reissuable: bool
        fee: int
        quantity: int
        decimals: int
        name: str
    """
    def __init__(self, sender: Account, name: str = None, reissuable: bool = None,
                 quantity: int = None, decimals: int = None, description: str = None,
                 timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        self.issue_data = drop_none({
            "name": name,
            "reissuable": reissuable,
            "quantity": quantity,
            "decimals": decimals,
            "description": description,
            "timestamp": timestamp,
            "fee": fee
        })
        super().__init__('Issue Token', self.issue_data)
        self.sender = sender
        self.history = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.issue.validators import validate_issue

        return validate_issue(self.sender, self.issue_data)
    

    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.issue.validators import mount_issue

        return super().transaction(
            mount_tx=mount_issue,
            sender=self.sender,
            issue_data=self.issue_data)


    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.issue.validators import send_issue

        tx = super().send(send_issue, node_url)
        self.history.append(tx)
        return tx


class IssueAsset(IssueToken):
    def __init__(self, sender: Account, name: str = None, reissuable: bool = None,
                 quantity: int = None, decimals: int = None, description: str = None,
                 timestamp: int = None, fee: int = None) -> None:
        from lunespy.utils import drop_none

        issue_data = drop_none({
            "name": name,
            "reissuable": reissuable,
            "quantity": quantity,
            "decimals": decimals,
            "description": description,
            "timestamp": timestamp,
            "fee": fee
        })
        IssueToken.__init__(self, sender=sender, **issue_data)
        BaseTransaction.__init__(self, tx_type='Issue Asset', tx_data=issue_data)


class IssueNFT(IssueToken):
    def __init__(self, sender: Account, name: str = None, reissuable: bool = None,
                 quantity: int = None, description: str = None, timestamp: int = None,
                 fee: int = None) -> None:
        from lunespy.utils import drop_none

        NFT_is_unique: int = 0
        issue_data = drop_none({
            "name": name,
            "reissuable": reissuable,
            "quantity": quantity,
            "decimals": NFT_is_unique,
            "description": description,
            "timestamp": timestamp,
            "fee": fee
        })
        IssueToken.__init__(self, sender=sender, **issue_data)
        BaseTransaction.__init__(self, tx_type='Issue NFT', tx_data=issue_data)



class WalletError(Exception):
    """Base error class for Wallet"""
    def __init__(self) -> None:
        super().__init__(self.__doc__)

class InvalidVersionAddress(WalletError):
    """Wrong address version"""   

class InvalidChainAddress(WalletError):
    """Wrong chain id"""

class InvalidLengthAddress(WalletError):
    """Wrong address length"""

class InvalidChecksumAddress(WalletError):
    """Wrong address checksum"""

class InvalidNonce(WalletError):
    """Nonce must be between 0 and 4294967295"""


class Account:
    """
    This is a class for Wallet Account.

    Parameters: 
        private_key(str): The pivate key of your wallet.
        public_key(str): The public key of your wallet.
        hash_seed(str): The seed of your wallet hashed.
        chain_id(str): The id of your chain. | '0' xor '1'
        address(str): The address of your wallet.
        nonce(int): The ip of your nonce. | 0 to 4.294.967.295
        chain(str): The chain of your node. | 'mainnet' xor 'testnet'
        seed(str): The seed of your wallet. | n words
        n_words(int): The amount words of your seed wallet. | n_words should be n // 3  
        byte_private_key(bytes): The byte of your wallet private key.
        byte_public_key(bytes): The byte of your wallet public key.
        byte_address(bytes): The byte of your wallet address
    """

    def __init__(self, n_words: int = None, seed: str = None, nonce: int = None,
                 chain: str = None, private_key: str = None, public_key: str = None,
                 hash_seed:str = None, chain_id: str = None, address: str = None) -> None:
        from lunespy.client.wallet.validators import validate_wallet
        from lunespy.utils import drop_none

        wallet: dict = drop_none({
            'private_key': private_key,
            'public_key': public_key,
            'hash_seed': hash_seed,
            'chain_id': chain_id,
            'address': address,
            'nonce': nonce,
            'chain': chain,
            'seed': seed,
            'n_words': n_words
        })

        self.__dict__ = validate_wallet(wallet)


    def __str__(self) -> str:
        from lunespy.utils import bcolors

        data = ''
        for key, value in self.__dict__.items():
            if 'byte' not in key and 'id' not in key:
                data += f"\n{key}{bcolors.OKGREEN}\n └── {value}{bcolors.ENDC}"
        return data


    def to_json(self, path: str = './data/') -> str:
        from lunespy.utils import export_json

        data = {
            key: value 
            for (key, value) in self.__dict__.items()
            if 'byte' not in key and 'id' not in key
        }

        return export_json(
            data=data,
            name=f'wallet-{self.chain}',
            path=path
        )


class LunexAccount(Account):
    """
    params: 
        public_key: str
        chain: str | 'mainnet' xor 'testnet'
        url: str | 'http://ip:port'
    """
    def __init__(self, **data):
        super().__init__(**data)
        self.url: str  = data['url']



from enum import Enum

from lunespy.utils import lunes_to_unes


class IssueType(Enum):
    to_int = 3
    to_byte = b'\x03'
    fee = lunes_to_unes(1)


class TransferType(Enum):
    to_int = 4
    to_byte = b'\x04'
    fee = lunes_to_unes(0.01)


class ReissueType(Enum):
    to_int = 5
    to_byte = b'\x05'
    fee = lunes_to_unes(0.01)


class BurnType(Enum):
    to_int = 6
    to_byte = b'\x06'
    fee = lunes_to_unes(0.01)


class LunexType(Enum):
    to_int = 7
    to_byte = b'\x07'
    fee = lunes_to_unes(0.01)


class LeaseType(Enum):
    to_int = 8
    to_byte = b'\x08'
    fee = lunes_to_unes(0.01)


class CancelLeaseType(Enum):
    to_int = 9
    to_byte = b'\x09'
    fee = lunes_to_unes(0.01)


class AliasType(Enum):
    to_int = 10
    to_byte = b'\x0a'
    mount = b'\x02'
    fee = lunes_to_unes(0.01)


class MassType(Enum):
    to_int = 11
    to_byte = b'\x0b'
    fee = lunes_to_unes(0.005)



from requests import get
from lunespy.server.nodes import Node


def block_from_height(height: int, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/blocks/at/{height}'
    else:
        full_url = f'{node_url}/blocks/at/{height}' # You have pass your node url with https or other contents
    
    response = get(full_url)
    
    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def range_block(start_block: int, end_block: int, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/blocks/seq/{start_block}/{end_block}'
    else:
        full_url = f'{node_url}/blocks/seq/{start_block}/{end_block}' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def last_block(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/blocks/last'
    else:
        full_url = f'{node_url}/blocks/last' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
                'status': 'error',
                'response': response.text
            }


def blocks_generated_by_specified_address(address: str, start_block: int, end_block: int, node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/blocks/address/{address}/{start_block}/{end_block}'
    else:
        full_url = f'{node_url}/blocks/address/{address}/{start_block}/{end_block}' # You have pass your node url with https or other contents
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }



ADDRESS_VERSION = 1
ADDRESS_CHECKSUM_LENGTH = 4
ADDRESS_HASH_LENGTH = 20
ADDRESS_LENGTH = 1 + 1 + ADDRESS_CHECKSUM_LENGTH + ADDRESS_HASH_LENGTH



word_list = ['abandon', 'ability', 'able', 'about', 'above', 'absent', 'absorb', 'abstract', 'absurd', 'abuse', 'access',
            'accident', 'account', 'accuse', 'achieve', 'acid', 'acoustic', 'acquire', 'across', 'act', 'action',
            'actor', 'actress', 'actual', 'adapt', 'add', 'addict', 'address', 'adjust', 'admit', 'adult', 'advance',
            'advice', 'aerobic', 'affair', 'afford', 'afraid', 'again', 'age', 'agent', 'agree', 'ahead', 'aim', 'air',
            'airport', 'aisle', 'alarm', 'album', 'alcohol', 'alert', 'alien', 'all', 'alley', 'allow', 'almost',
            'alone', 'alpha', 'already', 'also', 'alter', 'always', 'amateur', 'amazing', 'among', 'amount', 'amused',
            'analyst', 'anchor', 'ancient', 'anger', 'angle', 'angry', 'animal', 'ankle', 'announce', 'annual',
            'another', 'answer', 'antenna', 'antique', 'anxiety', 'any', 'apart', 'apology', 'appear', 'apple',
            'approve', 'april', 'arch', 'arctic', 'area', 'arena', 'argue', 'arm', 'armed', 'armor', 'army', 'around',
            'arrange', 'arrest', 'arrive', 'arrow', 'art', 'artefact', 'artist', 'artwork', 'ask', 'aspect', 'assault',
            'asset', 'assist', 'assume', 'asthma', 'athlete', 'atom', 'attack', 'attend', 'attitude', 'attract',
            'auction', 'audit', 'august', 'aunt', 'author', 'auto', 'autumn', 'average', 'avocado', 'avoid', 'awake',
            'aware', 'away', 'awesome', 'awful', 'awkward', 'axis', 'baby', 'bachelor', 'bacon', 'badge', 'bag',
            'balance', 'balcony', 'ball', 'bamboo', 'banana', 'banner', 'bar', 'barely', 'bargain', 'barrel', 'base',
            'basic', 'basket', 'battle', 'beach', 'bean', 'beauty', 'because', 'become', 'beef', 'before', 'begin',
            'behave', 'behind', 'believe', 'below', 'belt', 'bench', 'benefit', 'best', 'betray', 'better', 'between',
            'beyond', 'bicycle', 'bid', 'bike', 'bind', 'biology', 'bird', 'birth', 'bitter', 'black', 'blade', 'blame',
            'blanket', 'blast', 'bleak', 'bless', 'blind', 'blood', 'blossom', 'blouse', 'blue', 'blur', 'blush',
            'board', 'boat', 'body', 'boil', 'bomb', 'bone', 'bonus', 'book', 'boost', 'border', 'boring', 'borrow',
            'boss', 'bottom', 'bounce', 'box', 'boy', 'bracket', 'brain', 'brand', 'brass', 'brave', 'bread', 'breeze',
            'brick', 'bridge', 'brief', 'bright', 'bring', 'brisk', 'broccoli', 'broken', 'bronze', 'broom', 'brother',
            'brown', 'brush', 'bubble', 'buddy', 'budget', 'buffalo', 'build', 'bulb', 'bulk', 'bullet', 'bundle',
            'bunker', 'burden', 'burger', 'burst', 'bus', 'business', 'busy', 'butter', 'buyer', 'buzz', 'cabbage',
            'cabin', 'cable', 'cactus', 'cage', 'cake', 'call', 'calm', 'camera', 'camp', 'can', 'canal', 'cancel',
            'candy', 'cannon', 'canoe', 'canvas', 'canyon', 'capable', 'capital', 'captain', 'car', 'carbon', 'card',
            'cargo', 'carpet', 'carry', 'cart', 'case', 'cash', 'casino', 'castle', 'casual', 'cat', 'catalog', 'catch',
            'category', 'cattle', 'caught', 'cause', 'caution', 'cave', 'ceiling', 'celery', 'cement', 'census',
            'century', 'cereal', 'certain', 'chair', 'chalk', 'champion', 'change', 'chaos', 'chapter', 'charge',
            'chase', 'chat', 'cheap', 'check', 'cheese', 'chef', 'cherry', 'chest', 'chicken', 'chief', 'child',
            'chimney', 'choice', 'choose', 'chronic', 'chuckle', 'chunk', 'churn', 'cigar', 'cinnamon', 'circle',
            'citizen', 'city', 'civil', 'claim', 'clap', 'clarify', 'claw', 'clay', 'clean', 'clerk', 'clever', 'click',
            'client', 'cliff', 'climb', 'clinic', 'clip', 'clock', 'clog', 'close', 'cloth', 'cloud', 'clown', 'club',
            'clump', 'cluster', 'clutch', 'coach', 'coast', 'coconut', 'code', 'coffee', 'coil', 'coin', 'collect',
            'color', 'column', 'combine', 'come', 'comfort', 'comic', 'common', 'company', 'concert', 'conduct',
            'confirm', 'congress', 'connect', 'consider', 'control', 'convince', 'cook', 'cool', 'copper', 'copy',
            'coral', 'core', 'corn', 'correct', 'cost', 'cotton', 'couch', 'country', 'couple', 'course', 'cousin',
            'cover', 'coyote', 'crack', 'cradle', 'craft', 'cram', 'crane', 'crash', 'crater', 'crawl', 'crazy',
            'cream', 'credit', 'creek', 'crew', 'cricket', 'crime', 'crisp', 'critic', 'crop', 'cross', 'crouch',
            'crowd', 'crucial', 'cruel', 'cruise', 'crumble', 'crunch', 'crush', 'cry', 'crystal', 'cube', 'culture',
            'cup', 'cupboard', 'curious', 'current', 'curtain', 'curve', 'cushion', 'custom', 'cute', 'cycle', 'dad',
            'damage', 'damp', 'dance', 'danger', 'daring', 'dash', 'daughter', 'dawn', 'day', 'deal', 'debate',
            'debris', 'decade', 'december', 'decide', 'decline', 'decorate', 'decrease', 'deer', 'defense', 'define',
            'defy', 'degree', 'delay', 'deliver', 'demand', 'demise', 'denial', 'dentist', 'deny', 'depart', 'depend',
            'deposit', 'depth', 'deputy', 'derive', 'describe', 'desert', 'design', 'desk', 'despair', 'destroy',
            'detail', 'detect', 'develop', 'device', 'devote', 'diagram', 'dial', 'diamond', 'diary', 'dice', 'diesel',
            'diet', 'differ', 'digital', 'dignity', 'dilemma', 'dinner', 'dinosaur', 'direct', 'dirt', 'disagree',
            'discover', 'disease', 'dish', 'dismiss', 'disorder', 'display', 'distance', 'divert', 'divide', 'divorce',
            'dizzy', 'doctor', 'document', 'dog', 'doll', 'dolphin', 'domain', 'donate', 'donkey', 'donor', 'door',
            'dose', 'double', 'dove', 'draft', 'dragon', 'drama', 'drastic', 'draw', 'dream', 'dress', 'drift', 'drill',
            'drink', 'drip', 'drive', 'drop', 'drum', 'dry', 'duck', 'dumb', 'dune', 'during', 'dust', 'dutch', 'duty',
            'dwarf', 'dynamic', 'eager', 'eagle', 'early', 'earn', 'earth', 'easily', 'east', 'easy', 'echo', 'ecology',
            'economy', 'edge', 'edit', 'educate', 'effort', 'egg', 'eight', 'either', 'elbow', 'elder', 'electric',
            'elegant', 'element', 'elephant', 'elevator', 'elite', 'else', 'embark', 'embody', 'embrace', 'emerge',
            'emotion', 'employ', 'empower', 'empty', 'enable', 'enact', 'end', 'endless', 'endorse', 'enemy', 'energy',
            'enforce', 'engage', 'engine', 'enhance', 'enjoy', 'enlist', 'enough', 'enrich', 'enroll', 'ensure',
            'enter', 'entire', 'entry', 'envelope', 'episode', 'equal', 'equip', 'era', 'erase', 'erode', 'erosion',
            'error', 'erupt', 'escape', 'essay', 'essence', 'estate', 'eternal', 'ethics', 'evidence', 'evil', 'evoke',
            'evolve', 'exact', 'example', 'excess', 'exchange', 'excite', 'exclude', 'excuse', 'execute', 'exercise',
            'exhaust', 'exhibit', 'exile', 'exist', 'exit', 'exotic', 'expand', 'expect', 'expire', 'explain', 'expose',
            'express', 'extend', 'extra', 'eye', 'eyebrow', 'fabric', 'face', 'faculty', 'fade', 'faint', 'faith',
            'fall', 'false', 'fame', 'family', 'famous', 'fan', 'fancy', 'fantasy', 'farm', 'fashion', 'fat', 'fatal',
            'father', 'fatigue', 'fault', 'favorite', 'feature', 'february', 'federal', 'fee', 'feed', 'feel', 'female',
            'fence', 'festival', 'fetch', 'fever', 'few', 'fiber', 'fiction', 'field', 'figure', 'file', 'film',
            'filter', 'final', 'find', 'fine', 'finger', 'finish', 'fire', 'firm', 'first', 'fiscal', 'fish', 'fit',
            'fitness', 'fix', 'flag', 'flame', 'flash', 'flat', 'flavor', 'flee', 'flight', 'flip', 'float', 'flock',
            'floor', 'flower', 'fluid', 'flush', 'fly', 'foam', 'focus', 'fog', 'foil', 'fold', 'follow', 'food',
            'foot', 'force', 'forest', 'forget', 'fork', 'fortune', 'forum', 'forward', 'fossil', 'foster', 'found',
            'fox', 'fragile', 'frame', 'frequent', 'fresh', 'friend', 'fringe', 'frog', 'front', 'frost', 'frown',
            'frozen', 'fruit', 'fuel', 'fun', 'funny', 'furnace', 'fury', 'future', 'gadget', 'gain', 'galaxy',
            'gallery', 'game', 'gap', 'garage', 'garbage', 'garden', 'garlic', 'garment', 'gas', 'gasp', 'gate',
            'gather', 'gauge', 'gaze', 'general', 'genius', 'genre', 'gentle', 'genuine', 'gesture', 'ghost', 'giant',
            'gift', 'giggle', 'ginger', 'giraffe', 'girl', 'give', 'glad', 'glance', 'glare', 'glass', 'glide',
            'glimpse', 'globe', 'gloom', 'glory', 'glove', 'glow', 'glue', 'goat', 'goddess', 'gold', 'good', 'goose',
            'gorilla', 'gospel', 'gossip', 'govern', 'gown', 'grab', 'grace', 'grain', 'grant', 'grape', 'grass',
            'gravity', 'great', 'green', 'grid', 'grief', 'grit', 'grocery', 'group', 'grow', 'grunt', 'guard', 'guess',
            'guide', 'guilt', 'guitar', 'gun', 'gym', 'habit', 'hair', 'half', 'hammer', 'hamster', 'hand', 'happy',
            'harbor', 'hard', 'harsh', 'harvest', 'hat', 'have', 'hawk', 'hazard', 'head', 'health', 'heart', 'heavy',
            'hedgehog', 'height', 'hello', 'helmet', 'help', 'hen', 'hero', 'hidden', 'high', 'hill', 'hint', 'hip',
            'hire', 'history', 'hobby', 'hockey', 'hold', 'hole', 'holiday', 'hollow', 'home', 'honey', 'hood', 'hope',
            'horn', 'horror', 'horse', 'hospital', 'host', 'hotel', 'hour', 'hover', 'hub', 'huge', 'human', 'humble',
            'humor', 'hundred', 'hungry', 'hunt', 'hurdle', 'hurry', 'hurt', 'husband', 'hybrid', 'ice', 'icon', 'idea',
            'identify', 'idle', 'ignore', 'ill', 'illegal', 'illness', 'image', 'imitate', 'immense', 'immune',
            'impact', 'impose', 'improve', 'impulse', 'inch', 'include', 'income', 'increase', 'index', 'indicate',
            'indoor', 'industry', 'infant', 'inflict', 'inform', 'inhale', 'inherit', 'initial', 'inject', 'injury',
            'inmate', 'inner', 'innocent', 'input', 'inquiry', 'insane', 'insect', 'inside', 'inspire', 'install',
            'intact', 'interest', 'into', 'invest', 'invite', 'involve', 'iron', 'island', 'isolate', 'issue', 'item',
            'ivory', 'jacket', 'jaguar', 'jar', 'jazz', 'jealous', 'jeans', 'jelly', 'jewel', 'job', 'join', 'joke',
            'journey', 'joy', 'judge', 'juice', 'jump', 'jungle', 'junior', 'junk', 'just', 'kangaroo', 'keen', 'keep',
            'ketchup', 'key', 'kick', 'kid', 'kidney', 'kind', 'kingdom', 'kiss', 'kit', 'kitchen', 'kite', 'kitten',
            'kiwi', 'knee', 'knife', 'knock', 'know', 'lab', 'label', 'labor', 'ladder', 'lady', 'lake', 'lamp',
            'language', 'laptop', 'large', 'later', 'latin', 'laugh', 'laundry', 'lava', 'law', 'lawn', 'lawsuit',
            'layer', 'lazy', 'leader', 'leaf', 'learn', 'leave', 'lecture', 'left', 'leg', 'legal', 'legend', 'leisure',
            'lemon', 'lend', 'length', 'lens', 'leopard', 'lesson', 'letter', 'level', 'liar', 'liberty', 'library',
            'license', 'life', 'lift', 'light', 'like', 'limb', 'limit', 'link', 'lion', 'liquid', 'list', 'little',
            'live', 'lizard', 'load', 'loan', 'lobster', 'local', 'lock', 'logic', 'lonely', 'long', 'loop', 'lottery',
            'loud', 'lounge', 'love', 'loyal', 'lucky', 'luggage', 'lumber', 'lunar', 'lunch', 'luxury', 'lyrics',
            'machine', 'mad', 'magic', 'magnet', 'maid', 'mail', 'main', 'major', 'make', 'mammal', 'man', 'manage',
            'mandate', 'mango', 'mansion', 'manual', 'maple', 'marble', 'march', 'margin', 'marine', 'market',
            'marriage', 'mask', 'mass', 'master', 'match', 'material', 'math', 'matrix', 'matter', 'maximum', 'maze',
            'meadow', 'mean', 'measure', 'meat', 'mechanic', 'medal', 'media', 'melody', 'melt', 'member', 'memory',
            'mention', 'menu', 'mercy', 'merge', 'merit', 'merry', 'mesh', 'message', 'metal', 'method', 'middle',
            'midnight', 'milk', 'million', 'mimic', 'mind', 'minimum', 'minor', 'minute', 'miracle', 'mirror', 'misery',
            'miss', 'mistake', 'mix', 'mixed', 'mixture', 'mobile', 'model', 'modify', 'mom', 'moment', 'monitor',
            'monkey', 'monster', 'month', 'moon', 'moral', 'more', 'morning', 'mosquito', 'mother', 'motion', 'motor',
            'mountain', 'mouse', 'move', 'movie', 'much', 'muffin', 'mule', 'multiply', 'muscle', 'museum', 'mushroom',
            'music', 'must', 'mutual', 'myself', 'mystery', 'myth', 'naive', 'name', 'napkin', 'narrow', 'nasty',
            'nation', 'nature', 'near', 'neck', 'need', 'negative', 'neglect', 'neither', 'nephew', 'nerve', 'nest',
            'net', 'network', 'neutral', 'never', 'news', 'next', 'nice', 'night', 'noble', 'noise', 'nominee',
            'noodle', 'normal', 'north', 'nose', 'notable', 'note', 'nothing', 'notice', 'novel', 'now', 'nuclear',
            'number', 'nurse', 'nut', 'oak', 'obey', 'object', 'oblige', 'obscure', 'observe', 'obtain', 'obvious',
            'occur', 'ocean', 'october', 'odor', 'off', 'offer', 'office', 'often', 'oil', 'okay', 'old', 'olive',
            'olympic', 'omit', 'once', 'one', 'onion', 'online', 'only', 'open', 'opera', 'opinion', 'oppose',
            'option', 'orange', 'orbit', 'orchard', 'order', 'ordinary', 'organ', 'orient', 'original', 'orphan',
            'ostrich', 'other', 'outdoor', 'outer', 'output', 'outside', 'oval', 'oven', 'over', 'own', 'owner',
            'oxygen', 'oyster', 'ozone', 'pact', 'paddle', 'page', 'pair', 'palace', 'palm', 'panda', 'panel', 'panic',
            'panther', 'paper', 'parade', 'parent', 'park', 'parrot', 'party', 'pass', 'patch', 'path', 'patient',
            'patrol', 'pattern', 'pause', 'pave', 'payment', 'peace', 'peanut', 'pear', 'peasant', 'pelican', 'pen',
            'penalty', 'pencil', 'people', 'pepper', 'perfect', 'permit', 'person', 'pet', 'phone', 'photo', 'phrase',
            'physical', 'piano', 'picnic', 'picture', 'piece', 'pig', 'pigeon', 'pill', 'pilot', 'pink', 'pioneer',
            'pipe', 'pistol', 'pitch', 'pizza', 'place', 'planet', 'plastic', 'plate', 'play', 'please', 'pledge',
            'pluck', 'plug', 'plunge', 'poem', 'poet', 'point', 'polar', 'pole', 'police', 'pond', 'pony', 'pool',
            'popular', 'portion', 'position', 'possible', 'post', 'potato', 'pottery', 'poverty', 'powder', 'power',
            'practice', 'praise', 'predict', 'prefer', 'prepare', 'present', 'pretty', 'prevent', 'price', 'pride',
            'primary', 'print', 'priority', 'prison', 'private', 'prize', 'problem', 'process', 'produce', 'profit',
            'program', 'project', 'promote', 'proof', 'property', 'prosper', 'protect', 'proud', 'provide', 'public',
            'pudding', 'pull', 'pulp', 'pulse', 'pumpkin', 'punch', 'pupil', 'puppy', 'purchase', 'purity', 'purpose',
            'purse', 'push', 'put', 'puzzle', 'pyramid', 'quality', 'quantum', 'quarter', 'question', 'quick', 'quit',
            'quiz', 'quote', 'rabbit', 'raccoon', 'race', 'rack', 'radar', 'radio', 'rail', 'rain', 'raise', 'rally',
            'ramp', 'ranch', 'random', 'range', 'rapid', 'rare', 'rate', 'rather', 'raven', 'raw', 'razor', 'ready',
            'real', 'reason', 'rebel', 'rebuild', 'recall', 'receive', 'recipe', 'record', 'recycle', 'reduce',
            'reflect', 'reform', 'refuse', 'region', 'regret', 'regular', 'reject', 'relax', 'release', 'relief',
            'rely', 'remain', 'remember', 'remind', 'remove', 'render', 'renew', 'rent', 'reopen', 'repair', 'repeat',
            'replace', 'report', 'require', 'rescue', 'resemble', 'resist', 'resource', 'response', 'result', 'retire',
            'retreat', 'return', 'reunion', 'reveal', 'review', 'reward', 'rhythm', 'rib', 'ribbon', 'rice', 'rich',
            'ride', 'ridge', 'rifle', 'right', 'rigid', 'ring', 'riot', 'ripple', 'risk', 'ritual', 'rival', 'river',
            'road', 'roast', 'robot', 'robust', 'rocket', 'romance', 'roof', 'rookie', 'room', 'rose', 'rotate',
            'rough', 'round', 'route', 'royal', 'rubber', 'rude', 'rug', 'rule', 'run', 'runway', 'rural', 'sad',
            'saddle', 'sadness', 'safe', 'sail', 'salad', 'salmon', 'salon', 'salt', 'salute', 'same', 'sample', 'sand',
            'satisfy', 'satoshi', 'sauce', 'sausage', 'save', 'say', 'scale', 'scan', 'scare', 'scatter', 'scene',
            'scheme', 'school', 'science', 'scissors', 'scorpion', 'scout', 'scrap', 'screen', 'script', 'scrub', 'sea',
            'search', 'season', 'seat', 'second', 'secret', 'section', 'security', 'seed', 'seek', 'segment', 'select',
            'sell', 'seminar', 'senior', 'sense', 'sentence', 'series', 'service', 'session', 'settle', 'setup',
            'seven', 'shadow', 'shaft', 'shallow', 'share', 'shed', 'shell', 'sheriff', 'shield', 'shift', 'shine',
            'ship', 'shiver', 'shock', 'shoe', 'shoot', 'shop', 'short', 'shoulder', 'shove', 'shrimp', 'shrug',
            'shuffle', 'shy', 'sibling', 'sick', 'side', 'siege', 'sight', 'sign', 'silent', 'silk', 'silly', 'silver',
            'similar', 'simple', 'since', 'sing', 'siren', 'sister', 'situate', 'six', 'size', 'skate', 'sketch', 'ski',
            'skill', 'skin', 'skirt', 'skull', 'slab', 'slam', 'sleep', 'slender', 'slice', 'slide', 'slight', 'slim',
            'slogan', 'slot', 'slow', 'slush', 'small', 'smart', 'smile', 'smoke', 'smooth', 'snack', 'snake', 'snap',
            'sniff', 'snow', 'soap', 'soccer', 'social', 'sock', 'soda', 'soft', 'solar', 'soldier', 'solid',
            'solution', 'solve', 'someone', 'song', 'soon', 'sorry', 'sort', 'soul', 'sound', 'soup', 'source', 'south',
            'space', 'spare', 'spatial', 'spawn', 'speak', 'special', 'speed', 'spell', 'spend', 'sphere', 'spice',
            'spider', 'spike', 'spin', 'spirit', 'split', 'spoil', 'sponsor', 'spoon', 'sport', 'spot', 'spray',
            'spread', 'spring', 'spy', 'square', 'squeeze', 'squirrel', 'stable', 'stadium', 'staff', 'stage', 'stairs',
            'stamp', 'stand', 'start', 'state', 'stay', 'steak', 'steel', 'stem', 'step', 'stereo', 'stick', 'still',
            'sting', 'stock', 'stomach', 'stone', 'stool', 'story', 'stove', 'strategy', 'street', 'strike', 'strong',
            'struggle', 'student', 'stuff', 'stumble', 'style', 'subject', 'submit', 'subway', 'success', 'such',
            'sudden', 'suffer', 'sugar', 'suggest', 'suit', 'summer', 'sun', 'sunny', 'sunset', 'super', 'supply',
            'supreme', 'sure', 'surface', 'surge', 'surprise', 'surround', 'survey', 'suspect', 'sustain', 'swallow',
            'swamp', 'swap', 'swarm', 'swear', 'sweet', 'swift', 'swim', 'swing', 'switch', 'sword', 'symbol',
            'symptom', 'syrup', 'system', 'table', 'tackle', 'tag', 'tail', 'talent', 'talk', 'tank', 'tape', 'target',
            'task', 'taste', 'tattoo', 'taxi', 'teach', 'team', 'tell', 'ten', 'tenant', 'tennis', 'tent', 'term',
            'test', 'text', 'thank', 'that', 'theme', 'then', 'theory', 'there', 'they', 'thing', 'this', 'thought',
            'three', 'thrive', 'throw', 'thumb', 'thunder', 'ticket', 'tide', 'tiger', 'tilt', 'timber', 'time', 'tiny',
            'tip', 'tired', 'tissue', 'title', 'toast', 'tobacco', 'today', 'toddler', 'toe', 'together', 'toilet',
            'token', 'tomato', 'tomorrow', 'tone', 'tongue', 'tonight', 'tool', 'tooth', 'top', 'topic', 'topple',
            'torch', 'tornado', 'tortoise', 'toss', 'total', 'tourist', 'toward', 'tower', 'town', 'toy', 'track',
            'trade', 'traffic', 'tragic', 'train', 'transfer', 'trap', 'trash', 'travel', 'tray', 'treat', 'tree',
            'trend', 'trial', 'tribe', 'trick', 'trigger', 'trim', 'trip', 'trophy', 'trouble', 'truck', 'true',
            'truly', 'trumpet', 'trust', 'truth', 'try', 'tube', 'tuition', 'tumble', 'tuna', 'tunnel', 'turkey',
            'turn', 'turtle', 'twelve', 'twenty', 'twice', 'twin', 'twist', 'two', 'type', 'typical', 'ugly',
            'umbrella', 'unable', 'unaware', 'uncle', 'uncover', 'under', 'undo', 'unfair', 'unfold', 'unhappy',
            'uniform', 'unique', 'unit', 'universe', 'unknown', 'unlock', 'until', 'unusual', 'unveil', 'update',
            'upgrade', 'uphold', 'upon', 'upper', 'upset', 'urban', 'urge', 'usage', 'use', 'used', 'useful', 'useless',
            'usual', 'utility', 'vacant', 'vacuum', 'vague', 'valid', 'valley', 'valve', 'van', 'vanish', 'vapor',
            'various', 'vast', 'vault', 'vehicle', 'velvet', 'vendor', 'venture', 'venue', 'verb', 'verify', 'version',
            'very', 'vessel', 'veteran', 'viable', 'vibrant', 'vicious', 'victory', 'video', 'view', 'village',
            'vintage', 'violin', 'virtual', 'virus', 'visa', 'visit', 'visual', 'vital', 'vivid', 'vocal', 'voice',
            'void', 'volcano', 'volume', 'vote', 'voyage', 'wage', 'wagon', 'wait', 'walk', 'wall', 'walnut', 'want',
            'warfare', 'warm', 'warrior', 'wash', 'wasp', 'waste', 'water', 'wave', 'way', 'wealth', 'weapon', 'wear',
            'weasel', 'weather', 'web', 'wedding', 'weekend', 'weird', 'welcome', 'west', 'wet', 'whale', 'what',
            'wheat', 'wheel', 'when', 'where', 'whip', 'whisper', 'wide', 'width', 'wife', 'wild', 'will', 'win',
            'window', 'wine', 'wing', 'wink', 'winner', 'winter', 'wire', 'wisdom', 'wise', 'wish', 'witness', 'wolf',
            'woman', 'wonder', 'wood', 'wool', 'word', 'work', 'world', 'worry', 'worth', 'wrap', 'wreck', 'wrestle',
            'wrist', 'write', 'wrong', 'yard', 'year', 'yellow', 'you', 'young', 'youth', 'zebra', 'zero', 'zone',
            'zoo']

